import sys, os
import numpy as np
from matplotlib import pyplot as plt
from scipy.optimize import curve_fit

import BioNetInput, BioNetTrajectory

def persistanceLengthFunction(x, P):
	return np.exp(-x/P)
	
def persistanceLengthFunction2(x, P, x0):
	return np.exp(-(x - x0)/P)
	
def persistanceLengthGeneralFunction(x, P, y0, yf):

	return y0 + (yf - y0)*(1-np.exp(-x/P))
		
def calculatePersistanceLength(N, R, l, t, fname):

	# Total number of points to use on contour
	numPoints = 3 * (N - 1) + 1

	# Data arrays
	ls = np.array([0.0 for i in range(numPoints)])
	freq = np.array([0 for i in range(numPoints - 1)])
	dps = np.array([0.0 for i in range(numPoints - 1)])
	dps2 = np.array([0.0 for i in range(numPoints - 1)])
	dpserr = np.array([0.0 for i in range(numPoints - 1)])

	# Get (average) contours
	ls = np.array([0.0 for i in range(numPoints)])
	for i in range(t.numFrames):
		anl = 0.0
		ls[0] += anl
		anl += np.linalg.norm(t.frame[i].sPos[0][-1] - t.frame[i].pos[0])
		ls[1] += anl
		anl += np.linalg.norm(t.frame[i].sPos[1][0] - t.frame[i].sPos[0][-1])
		ls[2] += anl
		for j in range(1, N - 1):

			anl += np.linalg.norm(t.frame[i].pos[j] - t.frame[i].sPos[j][0])
			ls[3 * j] += anl
			anl += np.linalg.norm(t.frame[i].sPos[j][-1] - t.frame[i].pos[j])
			ls[3 * j + 1] += anl
			anl += np.linalg.norm(t.frame[i].sPos[j + 1][0] - t.frame[i].sPos[j][-1])
			ls[3 * j + 2] += anl

		ls[-1] = ls[-2] + np.linalg.norm(t.frame[i].pos[-1] - t.frame[i].sPos[-1][0])
	ls /= t.numFrames


	# Calculate persistences (using copies along the chain)
	for i in range(t.numFrames):

		v0 = t.frame[i].sPos[0][-1] - t.frame[i].pos[0]
		v0 /= np.linalg.norm(v0)

		jNIndex = -1
		for j in range(numPoints - 1):

			# Node index
			if j % 3 == 0:
				jNIndex += 1

			# Tangent vector
			if j % 3 == 0:
				vj = t.frame[i].sPos[jNIndex][-1] - t.frame[i].pos[jNIndex]
			elif j % 3 == 1:
				vj = t.frame[i].sPos[jNIndex + 1][0] - t.frame[i].sPos[jNIndex][-1]
			else:
				vj = t.frame[i].pos[jNIndex + 1] - t.frame[i].sPos[jNIndex + 1][0]

			vj /= np.linalg.norm(vj)

			# Cos angle is dot product
			d = np.dot(vj,v0)

			# Stats arrays
			dps[j] += d
			dps2[j] += d * d
			freq[j] += 1

	# Get actual values
	numRouseTimes = t.numFrames
	for i in range(numPoints - 1):
		dps[i] /= freq[i]
		dps2[i] /= freq[i]
		dpserr[i] = np.sqrt(dps2[i] - np.power(dps[i],2)) / np.sqrt(numRouseTimes)

	# Fix
	dpserr[0] = 0.01

	# Thin data to eliminate the heterogeneity
	ls = ls[:-1]#[::3]
#	dps = dps[::3]
#	dpserr = dpserr[::3]
#	numRouseTimes = numRouseTimes[::3]

	# Plot a curve?
	fig = plt.figure(1, figsize=(3.2, 3.2))
	plt.errorbar(ls,dps,yerr=dpserr, fmt='bx-', markersize=3, label="Raw Data")

	# Fit a curve
	sig = dpserr * np.sqrt(numRouseTimes)
	general = False
	if(general):
		func = persistanceLengthGeneralFunction
		guess = [100,1.0,0.0]
		bounds = ([-np.inf,0.99,-0.01], [np.inf,1.01,0.01])
	else:
		func = persistanceLengthFunction
		guess = [100]
		bounds = ([-np.inf], [np.inf])

	# Three fits
	ls2 = []
	ls2.extend(ls[2::3])
	ls2 = np.array(ls2)
	dps2 = []
	dps2.extend(dps[2::3])
	dps2 = np.array(dps2)
	sig2 = []
	sig2.extend(sig[2::3])
	sig2 = np.array(sig2)
	
	popt, pcov = curve_fit(func, ls[::3], dps[::3], guess, bounds=bounds, sigma=sig[::3])
	plt.plot(ls[::3], func(ls[::3], *popt), "--r", label=r"$l_P = (%5.3f \pm %5.3f)$nm" % (popt[0], np.sqrt(pcov[0][0])))

	# Get the normalised values as well
	Lp = popt[0]
	LpErr = np.sqrt(pcov[0][0])

	LcUnits = [l2 - l1 for l2,l1 in zip(ls[::3][1:], ls[::3][:-1])]
	LcUnit = np.mean(LcUnits)
	LcUnitErr = np.std(LcUnits) / np.sqrt(len(LcUnits))
	
	LpPrime = Lp / LcUnit
	LpPrimeErr = LpPrime * np.sqrt((LpErr/Lp)**2 + (LcUnitErr/LcUnit)**2)


	# Prettify
	plt.xlabel("|s' - s| (nm)", fontsize=MEDIUM)
	plt.ylabel(r"$\langle \vec{t(s)} \cdot \vec{t(0)} \rangle$", fontsize=MEDIUM)
	plt.legend(loc=0, fontsize=SMALL)
	ax00 = plt.gca()
	ax00.set_xticklabels(["%.1f" % bit for bit in ax00.get_xticks()], {"fontsize":MEDIUM})
	ax00.set_yticklabels(["%.1f" % bit for bit in ax00.get_yticks()], {"fontsize":MEDIUM})
	plt.tight_layout()
	plt.savefig("../results/" + fname + "_CorrelationFull.png", dpi=300)
	plt.clf()
	return Lp, LpErr, LpPrime, LpPrimeErr, popt[-1], np.sqrt(pcov[-1][-1]), (ls[::3], dps[::3], dpserr[::3], popt, func)

SMALL = 8
MEDIUM = 14
LARGE = 12

# System data
kbt = 4.11
monomersPerPolymer = 10
radius = 2.5		
drag = 6 * np.pi * radius

linkerLp = 0.4
linkerNumAA = 7
linkerL0 = 7 * 0.4

linkerL = 0
#linkerK = 2 * linkerLp * linkerL0
#linkerDl = np.sqrt(kbt / linkerK)

linkerDl = np.sqrt(2 * linkerLp * linkerL0)
linkerK = kbt / (linkerDl ** 2)

# Data filename
dFname = "../results/SinglePolymer_ParametrisedPersitenceLengths.dat"
fout = open(dFname, "w")
fout.write("Model, dlOn2R, Lp, LpErr,Lp',LpErr'\n")

# Variables
models = [[0,1,1,1,1,1,1,1,1,2]]
modelNames = ["GC%d-A" % (i) for i in range(len(models))]
lOn2Rs = [0.132,0.061,0.023]
lOn2RNames = ["Flexible", "SemiFlexible", "Rigid"]

# Master figure
masterfig = plt.figure(0, figsize=(3.2,3.2))

# Loop variables
lRIndex = 0
for lOn2R in lOn2Rs:

	# Calculate the new values, given that the polymer length needs to stay constant
	newRadius = (2 * radius + linkerDl) / (2 * (lOn2R + 1))
	newLinkerDl = 2 * newRadius * lOn2R
	newLinkerK = kbt / np.power(newLinkerDl, 2.0)
	
	mIndex = 0
	for model in models[:1]:

		bFname = "SinglePolymer_Model%s_%s" % (modelNames[mIndex], lOn2RNames[lRIndex])
		
		# Get objects
		iFname = "../input/" + bFname + ".xml"	
		inp = BioNetInput.BioNetInput()
		inp.load(iFname)
		
		tFname = "../output/" + bFname + ".bntrj"
		traj = BioNetTrajectory.BioNetTrajectory()
		traj.load(tFname, inScript=inp)
		
		p, perr, pPrime, pPrimeErr, yf, yferr, plottable = calculatePersistanceLength(monomersPerPolymer, newRadius, linkerL, traj, bFname)

		twoR = 2 * radius
		newN = 5
		Lc = twoR * ((lOn2R + 1) * (newN - 1) + 1)
		print("%s: dlOn2R = %f, k = %f, Lc = %f, Lp = %f, Lp / Lc = %f" % (lOn2RNames[lRIndex], lOn2R, kbt / ((lOn2R * twoR) ** 2), Lc, p, p / Lc))
		
		# Plot		
		plt.figure(masterfig.number)
		X = plottable[0]
		Y = plottable[1]
		YErr = plottable[2]
		
		index = 0
		for anX in X:
			if anX > 55:
				break
			index += 1
		X = X[:index]
		Y = Y[:index]
		YErr = YErr[:index]
		plt.errorbar(X,Y,yerr=YErr, fmt='C%dx' % (lRIndex), markersize=3, label=r"$\Delta l / 2R$ = %.2f" % (lOn2R))
		plt.plot(X, plottable[4](X, plottable[3]), "--C%d" % (lRIndex))

		# Data output
#		pPrime = p / (2 * newRadius + newLinkerDl + linkerL)
#		pPrimeErr = perr / (2 * newRadius + newLinkerDl + linkerL)
		
		fout.write("%s,%f,%f,%f,%f,%f\n" % (modelNames[mIndex], lOn2R, p, perr, pPrime, pPrimeErr))

	lRIndex += 1
		
fout.close()

	
# Prettify
plt.figure(masterfig.number)
plt.xlabel("|s' - s| (nm)", fontsize=MEDIUM)
plt.ylabel(r"$\langle \vec{t(s)} \cdot \vec{t(0)} \rangle$", fontsize=MEDIUM)
plt.legend(loc=0, fontsize=SMALL)
ax00 = plt.gca()
ax00.set_xlim(ax00.get_xlim()[0],60)
ax00.set_xticklabels(["%.1f" % bit for bit in ax00.get_xticks()], {"fontsize":MEDIUM})
ax00.set_yticklabels(["%.1f" % bit for bit in ax00.get_yticks()], {"fontsize":MEDIUM})
plt.tight_layout()
plt.savefig("../results/SinglePolymer_ParametrisedPersitenceLengthCorrelations.png", dpi=300)

